//
//  TupleRegex.swift
//  SwiftRegex4
//
//  Created by John Holdsworth on 13/01/2018.
//  Copyright © 2017 John Holdsworth. All rights reserved.
//
//  Repo: https://github.com/johnno1962/SwiftRegex5
//
//  $Id: //depot/SwiftRegex4/SwiftRegex5.playground/Sources/TupleRegex.swift#35 $
//

import Foundation

public protocol RegexLiteral {
    var regexPattern: String { get }
    var regexOptions: NSRegularExpression.Options { get }
}

struct RegexOptioned: RegexLiteral {
    public let regexPattern: String
    public let regexOptions: NSRegularExpression.Options
}

extension RegexLiteral {
    func regexOptions(add options: NSRegularExpression.Options) -> RegexLiteral {
        return RegexOptioned(regexPattern: regexPattern, regexOptions: regexOptions.union(options))
    }
    public var caseInsensitive: RegexLiteral {
        return regexOptions(add: [.caseInsensitive])
    }
    public var regexLazy: RegexLazy {
        return RegexLazy(literal: self)
    }
}

public struct RegexLazy {
    let literal: RegexLiteral
}

extension String: RegexLiteral {
    public static var regexDefaultOptions: NSRegularExpression.Options = []

    public var regexPattern: String {
        return self
    }
    public var regexOptions: NSRegularExpression.Options {
        return String.regexDefaultOptions
    }

    func nsRange(pos: Int = 0) -> NSRange {
        return NSMakeRange(pos, utf16.count-pos)
    }
    subscript(range: NSRange) -> Substring? {
        return Range(range, in: self).flatMap { self[$0] }
    }

    // regex literal aware compiler patterns
    public subscript<T>(regex: Regex<T>) -> T? {
        get {
            return regex.match(target: self)
        }
        set(newValue) {
            self = regex.replacing(target: self, template: newValue!)
        }
    }
    public subscript<T>(regex: Regex<T>.RegexAll) -> [T] {
        get {
            return regex.regex.matches(target: self)
        }
        set(newValues) {
            self = regex.regex.replacing(target: self, templates: newValues)
        }
    }

    public subscript<T>(regex: Regex<T>.RegexNext) -> AnyIterator<T> {
        return regex.regex.iterator(target: self)
    }
    public subscript<T>(regex: Regex<T>) -> (T, UnsafeMutablePointer<ObjCBool>) -> String {
        get {
            fatalError("Invalid get of closure")
        }
        set(closure) {
            self = regex.replacing(target: self, exec: closure)
        }
    }

    // plain old string regex patterns
    public subscript(pattern: RegexLiteral) -> Bool {
        return Regex<String>(pattern: pattern).matchResult(target: self) != nil
    }

    // single match and replace
    public subscript<T>(pattern: RegexLiteral) -> T? {
        get {
            return self[Regex<T>(pattern: pattern)]
        }
        set(newValue) {
            self[Regex<T>(pattern: pattern)] = newValue
        }
    }

    // mutliple match and replace
    public subscript<T>(pattern: RegexLiteral) -> [T] {
        get {
            return Regex<T>(pattern: pattern).matches(target: self)
        }
        set(newValues) {
            self = Regex<T>(pattern: pattern).replacing(target: self, templates: newValues)
        }
    }

    // operating on individual group
    public subscript(pattern: RegexLiteral, group: Int) -> String? {
        get {
            return Regex<[Substring?]>(pattern: pattern)
                .match(target: self)?[group].flatMap { String($0) }
        }
        set(newValue) {
            self = Regex<String>(pattern: pattern)
                .replacing(target: self, template: newValue!, group: group)
        }
    }

    // lazy iterators
    public subscript<T>(pattern: RegexLazy) -> AnyIterator<T> {
        return Regex<T>(pattern: pattern.literal).iterator(target: self)
    }

    // closure replace
    public subscript<T>(pattern: RegexLiteral) -> (T, UnsafeMutablePointer<ObjCBool>) -> String {
        get {
            fatalError("Invalid get of closure")
        }
        set(closure) {
            self = Regex<T>(pattern: pattern).replacing(target: self, exec: closure)
        }
    }
    public subscript<T>(pattern: RegexLiteral, group: Int) -> (T, UnsafeMutablePointer<ObjCBool>) -> String {
        get {
            fatalError("Invalid get of closure")
        }
        set(closure) {
            self = Regex<T>(pattern: pattern).replacing(target: self, group: group, exec: closure)
        }
    }

    // inplace replace
    public subscript(pattern: RegexLiteral, template: String) -> String {
        return Regex<String>(pattern: pattern).replacing(target: self, template: template)
    }
    public subscript(pattern: RegexLiteral, group: Int, template: String) -> String {
        return Regex<String>(pattern: pattern).replacing(target: self, template: template, group: group)
    }
    public subscript<T>(pattern: RegexLiteral, templates: [T]) -> String {
        return Regex<T>(pattern: pattern).replacing(target: self, templates: templates)
    }

    // switch/case
    public subscript<T>(match: RegexMatch) -> T {
        return Regex<T>(pattern: "").entuple(match: match.match!, from: self)
    }
}

private var regexCache = [UInt: [String: NSRegularExpression]]()

public final class Regex<T>: ExpressibleByStringLiteral {

    let literal: RegexLiteral

    lazy var regex: NSRegularExpression = {
        let pattern = literal.regexPattern, options = pattern.regexOptions
        do {
            if let regex = regexCache[options.rawValue]?[pattern] {
                return regex
            }
            let regex = try NSRegularExpression(pattern: pattern, options: options)
            if regexCache[options.rawValue] == nil {
                regexCache[options.rawValue] = [String: NSRegularExpression]()
            }
            regexCache[options.rawValue]![pattern] = regex
            return regex
        } catch {
            fatalError("Could not parse regex: \(pattern) - \(error)")
        }
    }()

    public init(pattern: RegexLiteral) {
        literal = pattern
    }
    public init(stringLiteral value: String) {
        literal = value
    }

    func matchResult(target: String) -> NSTextCheckingResult? {
        return regex.firstMatch(in: target, options: [], range: target.nsRange())
    }
    func entuple(match: NSTextCheckingResult, from target: String) -> T {
        if let match = match as? T {
            return match
        }

        let ranges = (0 ..< match.numberOfRanges).map { match.range(at: $0) }
        if let ranges = ranges as? T {
            return ranges
        }

        let substrs = ranges.map { target[$0] }
        if let substrs = substrs as? T {
            return substrs
        }

        let groups = substrs.map { String($0 ?? "") }
        if let groups = groups as? T {
            return groups
        }

#if true
        if let tuple = groups[0] as? T ?? substrs[0] as? T {
            return tuple
        }
        if let tuple =
                (groups[1], groups[2]) as? T ??
                (substrs[1], substrs[2]) as? T ??
                (groups[1], groups[2], groups[3]) as? T ??
                (substrs[1], substrs[2], substrs[3]) as? T ??
                (groups[1], groups[2], groups[3], groups[4]) as? T ??
                (substrs[1], substrs[2], substrs[3], substrs[4]) as? T {
            return tuple
        }
        if let tuple =
                (groups[1], groups[2], groups[3], groups[4], groups[5]) as? T ??
                (substrs[1], substrs[2], substrs[3], substrs[4], substrs[5]) as? T ??
                (groups[1], groups[2], groups[3], groups[4], groups[5], groups[6]) as? T ??
                (substrs[1], substrs[2], substrs[3], substrs[4], substrs[5], substrs[6]) as? T ??
                (groups[1], groups[2], groups[3], groups[4], groups[5], groups[6], groups[7]) as? T ??
                (substrs[1], substrs[2], substrs[3], substrs[4], substrs[5], substrs[6], substrs[7]) as? T {
            return tuple
        }
        if let tuple =
                (groups[1], groups[2], groups[3], groups[4], groups[5], groups[6], groups[7], groups[8]) as? T ??
                (substrs[1], substrs[2], substrs[3], substrs[4], substrs[5], substrs[6], substrs[7], substrs[8]) as? T ??
                (groups[1], groups[2], groups[3], groups[4], groups[5], groups[6], groups[7], groups[8], groups[9]) as? T ??
                (substrs[1], substrs[2], substrs[3], substrs[4], substrs[5], substrs[6], substrs[7], substrs[8], substrs[9]) as? T ??
                (groups[1], groups[2], groups[3], groups[4], groups[5], groups[6], groups[7], groups[8], groups[9], groups[10]) as? T ??
                (substrs[1], substrs[2], substrs[3], substrs[4], substrs[5], substrs[6], substrs[7], substrs[8], substrs[9], substrs[10]) as? T {
                return tuple
        }
#else
        // the alternative...
        return groups[(T.self == String.self ? 0 : 1)...].withUnsafeBufferPointer {
            $0.baseAddress!.withMemoryRebound(to: T.self, capacity: 1) { $0.pointee }
        }
#endif
        fatalError("Cannot extract type \(T.self)")
    }
    public func match(target: String) -> T? {
        return matchResult(target: target).flatMap { entuple(match: $0, from: target) }
    }
    public func matches(target: String) -> [T] {
        var out = [T]()
        regex.enumerateMatches(in: target, options: [], range: target.nsRange()) {
            (match: NSTextCheckingResult?, flags: NSRegularExpression.MatchingFlags, stop: UnsafeMutablePointer<ObjCBool>) in
            out.append(entuple(match: match!, from: target))
        }
        return out
    }

    func detuple(newValue: T) -> [String?] {
        let children = Mirror(reflecting: newValue).children
        return children.count != 0 ? children.map { $0.value as? String } : [newValue as? String]
    }
    func groupRange(match: NSTextCheckingResult, replacements: [String?]) -> CountableRange<Int> {
        return replacements.count == 1 ? 0 ..< 1 :
            (match.numberOfRanges == 1 ? 0 : 1) ..< min(match.numberOfRanges, replacements.count+1)
    }
    public func replacing(target: String, template: T, group: Int? = nil) -> String {
        return replacing(target: target, templates: [template], group: group, firstOnly: true)
    }
    public func replacing(target: String, templates: [T], group tgroup: Int? = nil, firstOnly: Bool = false) -> String {
        var out = [Substring]()
        var pos = 0, matchno = 0

        regex.enumerateMatches(in: target, options: [], range: target.nsRange()) {
            (match: NSTextCheckingResult?, flags: NSRegularExpression.MatchingFlags, stop: UnsafeMutablePointer<ObjCBool>) in
            guard let match = match else { return }
            let replacements = detuple(newValue: templates.count > 1 ? templates[matchno] : templates[0])

            matchno += 1
            if firstOnly || templates.count > 1 && matchno == templates.count {
                stop.pointee = true
            }

            for group in tgroup != nil ? tgroup! ..< tgroup!+1 :
                groupRange(match: match, replacements: replacements) {
                let range = match.range(at: group)
                if range.location != NSNotFound && range.location >= pos {
                    out.append(target[NSMakeRange(pos, range.location - pos)] ?? "Invalid range 1")
                    let replacement = replacements[tgroup != nil ? 0 : max(0,group-1)]
                    out.append(Substring(regex.replacementString(for: match, in: target, offset: 0,
                                                    template: replacement ?? String(target[range]!))))
                    pos = NSMaxRange(range)
                }
            }
        }

        out.append(target[target.nsRange(pos: pos)] ?? "Invalid range 2")
        return out.joined()
    }
    public func replacing(target: String, group: Int = 0, exec closure: (T, UnsafeMutablePointer<ObjCBool>) -> String) -> String {
        var out = [Substring]()
        var pos = 0

        regex.enumerateMatches(in: target, options: [], range: target.nsRange()) {
            (match: NSTextCheckingResult?, flags: NSRegularExpression.MatchingFlags, stop: UnsafeMutablePointer<ObjCBool>) in
            guard let match = match else { return }
            let range = match.range(at: group)
            out.append(target[NSMakeRange(pos, range.location - pos)] ?? "Invalid range 3")
            out.append(Substring(closure(entuple(match: match, from: target), stop)))
            pos = NSMaxRange(range)
        }

        out.append(target[target.nsRange(pos: pos)] ?? "Invalid range 4")
        return out.joined()
    }

    private struct TupleIterator: IteratorProtocol {
        public typealias Element = T
        let regex: Regex<T>
        let target: String
        var pos = 0

        init(regex: Regex<T>, target: String) {
            self.regex = regex
            self.target = target
        }

        public mutating func next() -> T? {
            if let match = regex.regex.firstMatch(in: target, options: [], range: target.nsRange(pos: pos)) {
                pos = NSMaxRange(match.range)
                return regex.entuple(match: match, from: target)
            }
            return nil
        }
    }
    public func iterator(target: String) -> AnyIterator<T> {
        var iterator = TupleIterator(regex: self, target: target)
        return AnyIterator {
            iterator.next()
        }
    }

    // alternative usage
    public var caseInsensitive: Regex<T> {
        return Regex<T>(pattern: literal.caseInsensitive)
    }
    public var allMatches: RegexAll {
        return RegexAll(regex: self)
    }
    public var iterate: RegexNext {
        return RegexNext(regex: self)
    }
    public struct RegexAll {
        let regex: Regex<T>
    }
    public struct RegexNext {
        let regex: Regex<T>
    }
}

public class RegexMatch {
    var match: NSTextCheckingResult?
    public init() {
    }
}

extension RegexLiteral {
    public func regex(capture: RegexMatch?) -> RegexPattern {
        return RegexPattern(literal: self, capture: capture)
    }
}

public struct RegexPattern {
    let literal: RegexLiteral
    let capture: RegexMatch?
    public static func ~= (pattern: RegexPattern, value: String) -> Bool {
        if let match = Regex<String>(pattern: pattern.literal).matchResult(target: value) {
            pattern.capture?.match = match
            return true
        }
        return false
    }
}
